#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "simpletools.h"
#include "config.h"
#include "debug.h"
#include "gcode_struct.h"
#include "adafruit_ads1015.h"
#include "sergcode.h"

static char fileline[100];

static int calc_steps_required(double distance, int32_t steps_per_mm)
{
	double steps;

	steps = distance * (double)steps_per_mm;

	return (uint32_t)floor(steps + 0.5);
}

static void perform_gcode_struct_calcs()
{
	if(current_gcode.X_NUM != 0 && current_gcode.X_NUM > previous_gcode.X_NUM)
	{
		current_gcode.X_DISTANCE = current_gcode.X_NUM - previous_gcode.X_NUM;
		current_gcode.X_DIR = X_INC_DIR;
		current_gcode.X_STEPS = 
			calc_steps_required(current_gcode.X_DISTANCE, X_STEPS_PER_MM);
	}     
	else if(current_gcode.X_NUM != 0 && current_gcode.X_NUM < previous_gcode.X_NUM)
	{
		current_gcode.X_DISTANCE = previous_gcode.X_NUM - current_gcode.X_NUM;
		current_gcode.X_DIR = X_DEC_DIR;
		current_gcode.X_STEPS = 
			calc_steps_required(current_gcode.X_DISTANCE, X_STEPS_PER_MM);
	} 
	else
	{
		current_gcode.X_DISTANCE = 0;
		current_gcode.X_DIR = previous_gcode.X_DIR;
		current_gcode.X_STEPS = 0;
	}

	if(current_gcode.Y_NUM != 0 && current_gcode.Y_NUM > previous_gcode.Y_NUM)
	{
		current_gcode.Y_DISTANCE = current_gcode.Y_NUM - previous_gcode.Y_NUM;
		current_gcode.Y_DIR = Y_INC_DIR;
		current_gcode.Y_STEPS = 
			calc_steps_required(current_gcode.Y_DISTANCE, Y_STEPS_PER_MM);
	}     
	else if(current_gcode.Y_NUM != 0 && current_gcode.Y_NUM < previous_gcode.Y_NUM)
	{
		current_gcode.Y_DISTANCE = previous_gcode.Y_NUM - current_gcode.Y_NUM;
		current_gcode.Y_DIR = Y_DEC_DIR;
		current_gcode.Y_STEPS = 
			calc_steps_required(current_gcode.Y_DISTANCE, Y_STEPS_PER_MM);
	} 
	else
	{
		current_gcode.Y_DISTANCE = 0;
		current_gcode.Y_DIR = previous_gcode.Y_DIR;
		current_gcode.Y_STEPS = 0;
	}

	if(current_gcode.Z_NUM != 0 && current_gcode.Z_NUM > previous_gcode.Z_NUM)
	{
		current_gcode.Z_DISTANCE = current_gcode.Z_NUM - previous_gcode.Z_NUM;
		current_gcode.Z_DIR = Z_INC_DIR;
		current_gcode.Z_STEPS = 
			calc_steps_required(current_gcode.Z_DISTANCE, Z_STEPS_PER_MM);
	}     
	else if(current_gcode.Z_NUM != 0 && current_gcode.Z_NUM < previous_gcode.Z_NUM)
	{
		current_gcode.Z_DISTANCE = previous_gcode.Z_NUM - current_gcode.Z_NUM;
		current_gcode.Z_DIR = Z_DEC_DIR;
		current_gcode.Z_STEPS = 
			calc_steps_required(current_gcode.Z_DISTANCE, Z_STEPS_PER_MM);
	} 
	else
	{
		current_gcode.Z_DISTANCE = 0;
		current_gcode.Z_DIR = previous_gcode.Z_DIR;
		current_gcode.Z_STEPS = 0;
	}

	if(current_gcode.E_NUM != 0 && current_gcode.E_NUM > previous_gcode.E_NUM)
	{
		current_gcode.E_DISTANCE = current_gcode.E_NUM - previous_gcode.E_NUM;
		current_gcode.E_DIR = E_INC_DIR;
		current_gcode.E_STEPS = 
			calc_steps_required(current_gcode.E_DISTANCE, E_STEPS_PER_MM);
	}     
	else if(current_gcode.E_NUM != 0 && current_gcode.E_NUM < previous_gcode.E_NUM)
	{
		current_gcode.E_DISTANCE = previous_gcode.E_NUM - current_gcode.E_NUM;
		current_gcode.E_DIR = E_DEC_DIR;
		current_gcode.E_STEPS = 
			calc_steps_required(current_gcode.E_DISTANCE, E_STEPS_PER_MM);
	} 
	else
	{
		current_gcode.E_DISTANCE = 0;
		current_gcode.E_DIR = previous_gcode.E_DIR;
		current_gcode.E_STEPS = 0;
	}
}

static void ini_gcode_struct()
{
	uint8_t gcode_char;
	uint8_t gcode_struct_element;
	uint8_t fileline_length;
	uint8_t value_counter;

	uint32_t counter;

	// Clear any lingering data
	memset(current_gcode.gcode_fileline, 0, strlen(current_gcode.gcode_fileline));
	memset(current_gcode.Valid, 0, strlen(current_gcode.Valid));
	memset(current_gcode.G_CHAR, 0, strlen(current_gcode.G_CHAR));
	memset(current_gcode.M_CHAR, 0, strlen(current_gcode.M_CHAR));
	memset(current_gcode.T_CHAR, 0, strlen(current_gcode.T_CHAR));
	memset(current_gcode.S_CHAR, 0, strlen(current_gcode.S_CHAR));
	memset(current_gcode.P_CHAR, 0, strlen(current_gcode.P_CHAR));
	memset(current_gcode.X_CHAR, 0, strlen(current_gcode.X_CHAR));
	memset(current_gcode.Y_CHAR, 0, strlen(current_gcode.Y_CHAR));
	memset(current_gcode.Z_CHAR, 0, strlen(current_gcode.Z_CHAR));
	memset(current_gcode.E_CHAR, 0, strlen(current_gcode.E_CHAR));
	memset(current_gcode.F_CHAR, 0, strlen(current_gcode.F_CHAR));

	memcpy(&current_gcode.gcode_fileline, &fileline, sizeof(fileline));

	// Get the length of the fileline that is being parsed
	fileline_length = (uint8_t)strlen(fileline);

	// Fill current_gcode.Valid with literal zeros
	for(counter = 0; counter < GCODE_ELEMENTS; counter++)
	{
		current_gcode.Valid[counter] = '0';
	}

	value_counter = 0;

	//Parse file line into seperate field values.  Various fields will remain empty
	for(counter = 0; counter < fileline_length; counter++)
	{
		gcode_char = fileline[counter];

		// Exclude these characters from becoming part of are field entries
		if(gcode_char == 13 /*Carriage Return*/ || gcode_char == 10 /*Line Feed*/)
		{
			// Break free from this loop
			break;
		}

		else if(gcode_char == ' ')
		{
			value_counter = 0;

			// Skip the remaining iterations, because
			// we found a match, however, stay in the loop,
			// because we are not done yet
			continue;
		}

		else if(gcode_char == 'G')
		{
			// Set current field element
			gcode_struct_element = 0;
			current_gcode.Valid[0] = '1';

			continue;
		}
		else if(gcode_char == 'M')
		{
			// Set current field element
			gcode_struct_element = 1;
			current_gcode.Valid[1] = '1';

			continue;
		}
		else if(gcode_char == 'T')
		{
			// Set current field element
			gcode_struct_element = 2;
			current_gcode.Valid[2] = '1';

			continue;
		}
		else if(gcode_char == 'S')
		{
			// Set current field element
			gcode_struct_element = 3;
			current_gcode.Valid[3] = '1';

			continue;
		}
		else if(gcode_char == 'P')
		{
			// Set current field element
			gcode_struct_element = 4;
			current_gcode.Valid[4] = '1';

			continue;
		}
		else if(gcode_char == 'X')
		{
			// Set current field element
			gcode_struct_element = 5;
			current_gcode.Valid[5] = '1';

			continue;
		}
		else if(gcode_char == 'Y')
		{
			// Set current field element
			gcode_struct_element = 6;
			current_gcode.Valid[6] = '1';

			continue;
		}
		else if(gcode_char == 'Z')
		{
			// Set current field element
			gcode_struct_element = 7;
			current_gcode.Valid[7] = '1';

			continue;
		}		
		else if(gcode_char == 'E')
		{
			// Set current field element
			gcode_struct_element = 8;
			current_gcode.Valid[8] = '1';

			continue;
		}
		else if(gcode_char == 'F')
		{
			// Set current field element
			gcode_struct_element = 9;
			current_gcode.Valid[9] = '1';

			continue;
		}
		else
		{
			if(gcode_struct_element == 0)
			{
				// G Field element character
				current_gcode.G_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 1)
			{
				// M Field element character
				current_gcode.M_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 2)
			{
				// T Field element character
				current_gcode.T_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 3)
			{
				// S Field element character
				current_gcode.S_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 4)
			{
				// P Field element character
				current_gcode.P_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 5)
			{
				// X Field element character
				current_gcode.X_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 6)
			{
				// Y Field element character
				current_gcode.Y_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 7)
			{
				// Z Field element character
				current_gcode.Z_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 8)
			{
				// E Field element character
				current_gcode.E_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}
			else if(gcode_struct_element == 9)
			{
				// F Field element character
				current_gcode.F_CHAR[value_counter] = gcode_char;
				value_counter++;

				continue;
			}			
		}
	}

	current_gcode.G_NUM = atoi(current_gcode.G_CHAR);
	current_gcode.M_NUM = atoi(current_gcode.M_CHAR);
	current_gcode.T_NUM = atoi(current_gcode.T_CHAR);
	current_gcode.S_NUM = atoi(current_gcode.S_CHAR);
	current_gcode.P_NUM = atoi(current_gcode.P_CHAR);

	current_gcode.X_NUM = atof(current_gcode.X_CHAR);
	current_gcode.Y_NUM = atof(current_gcode.Y_CHAR);
	current_gcode.Z_NUM = atof(current_gcode.Z_CHAR);
	current_gcode.E_NUM = atof(current_gcode.E_CHAR);
	current_gcode.F_NUM = atof(current_gcode.F_CHAR);

	perform_gcode_struct_calcs();
}

int main(void)
{
	adc_init();
//	serial_init();

	uint32_t line_counter = 0;

	// Set debug level
	debug_flags |= DEBUG_ERROR;

	// Initial previous_gcode GCODE_STRUCT members to zero
	previous_gcode.G_NUM = 0;
	previous_gcode.M_NUM = 0;
	previous_gcode.T_NUM = 0;
	previous_gcode.S_NUM = 0;
	previous_gcode.P_NUM = 0;
	previous_gcode.X_NUM = 0;
	previous_gcode.Y_NUM = 0;
	previous_gcode.Z_NUM = 0;
	previous_gcode.E_NUM = 0;
	previous_gcode.F_NUM = 0;

	previous_gcode.X_DISTANCE = 0;
	previous_gcode.Y_DISTANCE = 0;
	previous_gcode.Z_DISTANCE = 0;
	previous_gcode.E_DISTANCE = 0;

	previous_gcode.X_STEPS = 0;
	previous_gcode.Y_STEPS = 0;
	previous_gcode.Z_STEPS = 0;
	previous_gcode.E_STEPS = 0;

	previous_gcode.X_DIR = 0;
	previous_gcode.Y_DIR = 0;
	previous_gcode.Z_DIR = 0;
	previous_gcode.E_DIR = 0;

	// Verify that SD card pins are defined
	#if defined DO && defined CLK && defined DI && defined SD_CS

	// Mount SD card
	sd_mount(DO, CLK, DI, SD_CS);

	FILE * pFile;

	// Open up a file on the SD card
	pFile = fopen("DEBUG.TXT","r");

	if(pFile == NULL)
	{
		if(DEBUG_ERROR && (debug_flags & DEBUG_ERROR))
		{
			print("ERROR - The file could not be opened!");
		}
	}
	else
	{
		// Recursively grab filelines
		while(fgets(fileline, 100, pFile))
		{
			// Count the filelines as we go along
			line_counter++;

			// Exclude lines that begin with certain characters
			// E.G. Remove newlines, carriage returns, and lines
			// that begin with a semi-colon.
			if(fileline[0] != 59 && fileline[0] != 13 && fileline[0] != 10)
			{
				// Partially fill the GCODE_STRUCT with the data
				// obtained from the fileline
				ini_gcode_struct();

				// Trim off newline character from the fileline
				// returned by fgets
				fileline[strcspn(fileline, "\n")] = 0;

    				print("Fileline: %s\n", fileline);

    				print("Valid GCODE Fields: %s\n", current_gcode.Valid);

				// Check the various GCODE_STRUCT field lengths
				// to see if they contain data, and if so, then
				// print the data
				if((uint8_t)strlen(current_gcode.G_CHAR) != 0)
				{
					print("G: %s, ", current_gcode.G_CHAR);
				}

				if((uint8_t)strlen(current_gcode.M_CHAR) != 0)
				{
					print("M: %s, ", current_gcode.M_CHAR);
				}

				if((uint8_t)strlen(current_gcode.T_CHAR) != 0)
				{
					print("T: %s, ", current_gcode.T_CHAR);
				}

				if((uint8_t)strlen(current_gcode.S_CHAR) != 0)
				{
					print("S: %s, ", current_gcode.S_CHAR);
				}

				if((uint8_t)strlen(current_gcode.P_CHAR) != 0)
				{
					print("P: %s, ", current_gcode.P_CHAR);
				}

				if((uint8_t)strlen(current_gcode.X_CHAR) != 0)
				{
					print("X: %s, ", current_gcode.X_CHAR);
				}

				if((uint8_t)strlen(current_gcode.Y_CHAR) != 0)
				{
					print("Y: %s, ", current_gcode.Y_CHAR);
				}

				if((uint8_t)strlen(current_gcode.Z_CHAR) != 0)
				{
					print("Z: %s, ", current_gcode.Z_CHAR);
				}

				if((uint8_t)strlen(current_gcode.E_CHAR) != 0)
				{
					print("E: %s, ", current_gcode.E_CHAR);
				}

				if((uint8_t)strlen(current_gcode.F_CHAR) != 0)
				{
					print("F: %s, ", current_gcode.F_CHAR);
				}

				print("\n");

				if((uint8_t)strlen(current_gcode.G_CHAR) != 0)
				{
					print("G: %u, ", current_gcode.G_NUM);
				}

				if((uint8_t)strlen(current_gcode.M_CHAR) != 0)
				{
					print("M: %u, ", current_gcode.M_NUM);
				}

				if((uint8_t)strlen(current_gcode.T_CHAR) != 0)
				{
					print("T: %u, ", current_gcode.T_NUM);
				}

				if((uint8_t)strlen(current_gcode.S_CHAR) != 0)
				{
					print("S: %u, ", current_gcode.S_NUM);
				}

				if((uint8_t)strlen(current_gcode.P_CHAR) != 0)
				{
					print("P: %u, ", current_gcode.P_NUM);
				}

				if((uint8_t)strlen(current_gcode.X_CHAR) != 0)
				{
					print("X: %5.2f, ", current_gcode.X_NUM);
				}

				if((uint8_t)strlen(current_gcode.Y_CHAR) != 0)
				{
					print("Y: %5.2f, ", current_gcode.Y_NUM);
				}

				if((uint8_t)strlen(current_gcode.Z_CHAR) != 0)
				{
					print("Z: %5.2f, ", current_gcode.Z_NUM);
				}

				if((uint8_t)strlen(current_gcode.E_CHAR) != 0)
				{
					print("E: %5.2f, ", current_gcode.E_NUM);
				}

				if((uint8_t)strlen(current_gcode.F_CHAR) != 0)
				{
					print("F: %5.2f, ", current_gcode.F_NUM);
				}

				print("\n");

				print("Steps Required: X: %u, Y: %u, Z: %u, E: %u\n",
					current_gcode.X_STEPS, current_gcode.Y_STEPS,
					current_gcode.Z_STEPS, current_gcode.E_STEPS);

				print("Distance: X: %5.2f, Y: %5.2f, Z: %5.2f, E: %5.2f\n",
					current_gcode.X_DISTANCE, current_gcode.Y_DISTANCE,
					current_gcode.Z_DISTANCE, current_gcode.E_DISTANCE);

				print("Motor Rotation: X: %u, Y: %u, Z: %u, E: %u\n",
					current_gcode.X_DIR, current_gcode.Y_DIR,
					current_gcode.Z_DIR, current_gcode.E_DIR);

				print("Previous Values: G: %u, ", previous_gcode.G_NUM);
				print("M: %u, ", previous_gcode.M_NUM);
				print("T: %u, ", previous_gcode.T_NUM);
				print("S: %u, ", previous_gcode.S_NUM);
				print("P: %u, ", previous_gcode.P_NUM);
				print("X: %5.2f, ", previous_gcode.X_NUM);
				print("Y: %5.2f, ", previous_gcode.Y_NUM);
				print("Z: %5.2f, ", previous_gcode.Z_NUM);
				print("E: %5.2f, ", previous_gcode.E_NUM);
				print("F: %5.2f, ", previous_gcode.F_NUM);
				print("\nPrevious Steps Required: X: %u, Y: %u, Z: %u, E: %u",
					previous_gcode.X_STEPS, previous_gcode.Y_STEPS,
					previous_gcode.Z_STEPS, previous_gcode.E_STEPS);

				print("\n\n");
			}

			// Copy the current GCODE_STRUCT into the previous one
			memcpy(&previous_gcode, &current_gcode, sizeof(GCODE_STRUCT));

			// Clear any lingering data from the fileline variable
			memset(fileline, 0, strlen(fileline));
		}

		// Close up the file on the SD card
		fclose(pFile);

		// Print the number of filelines encountered
		print("%u total lines", line_counter);
	}
 
	#else
    
	if(DEBUG_ERROR && (debug_flags & DEBUG_ERROR))
	{
		print("ERROR - The proper inputs and outputs must be defined for SD card usage!");
	}
    
	#endif
    
	return 0;
}